/******************************************************************************
*
*   This file is part of the software tool "Cervix Image Processing 2 (CIP2)".
*
*   Author: Christoph Bergmeir
*   Supervisor: José M. Benítez
*   Copyright (c) DiCITS Lab, Sci2s group, DECSAI, University of Granada.
*
*   This program is free software; you can redistribute it and/or
*   modify it under the terms of the GNU General Public License
*   as published by the Free Software Foundation; either version 2
*   of the License, or (at your option) any later version.
*   
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
*   MA  02110-1301, USA.
*
******************************************************************************/

#include <iostream>
#include <sstream>

#include <cv.h>
#include <highgui.h>
#include <math.h>

static const double PI = 3.14159265358979323846;

typedef struct MyLine
{
  float vx,vy,px,py;
}
MyLine;

CV_INLINE  MyLine  myLine( float vx, float vy, float px, float py )
{
  MyLine l;

  l.vx = vx;
  l.vy = vy;
  l.px = px;
  l.py = py;

  return l;
}

CV_INLINE  CvPoint2D32f  intersect( MyLine* l1, MyLine* l2)
{
  CvPoint2D32f t;

  float num = l1->vx*(l2->py-l1->py) + l1->vy*(l1->px-l2->px);
  float denom = l1->vy*l2->vx - l1->vx*l2->vy;
  float mu = num/denom;

  t.x = l2->vx * mu + l2->px;
  t.y = l2->vy * mu + l2->py;

  return t;
}

int main(int argc, char** argv)
{
  int roi_size = atoi( argv[3] );
  int roi_x = atoi( argv[4] );
  int roi_y = atoi( argv[5] );
  int samples_num = atoi( argv[6] );
  int kernel_size = atoi( argv[7] );
  float min_dist = (float) atof( argv[8] );
  float max_dist = (float) atof( argv[9] );

  int verbosity = 0;
  if (argc>9) verbosity = atoi( argv[10] );

  int thres_preproc = 150;
  int kernel_num_points = 10;

  //define number of clusters
  int cluster_count = 5;

  //define colors for the different ellipse clusters
  CvScalar color_tab[5];

  color_tab[0] = CV_RGB(255,0,0);
  color_tab[1] = CV_RGB(0,255,0);
  color_tab[2] = CV_RGB(100,100,255);
  color_tab[3] = CV_RGB(255,0,255);
  color_tab[4] = CV_RGB(255,255,0);

  IplImage* src_image = cvLoadImage(argv[1], 0);
  if(src_image == NULL) 
  {
    std::cout << "Source image not found\n";
    return 0;
  }

  IplImage* roi = cvCreateImage( cvSize(roi_size,roi_size), 8, 1 );
  cvSetImageROI(src_image,cvRect(roi_x,roi_y,roi_size,roi_size)); 
  cvCopy(src_image,roi); 
  cvResetImageROI(src_image);

  IplImage* color_roi = cvCreateImage( cvGetSize(roi), 8, 3 );
  cvCvtColor( roi, color_roi, CV_GRAY2BGR );

  CvMemStorage* all_points_storage = cvCreateMemStorage(0);
  CvSeq* all_points_seq = cvCreateSeq( CV_32FC2, sizeof(CvSeq), sizeof(CvPoint2D32f), all_points_storage );

  int nl= roi->height;
  int nc= roi->width;
  int step= roi->widthStep;

  unsigned char *roi_data= reinterpret_cast<unsigned char *>(roi->imageData);

  //put all points in the roi with a value greater thres_preproc into the array for further processing
  int num_all_points = 0;
  for (int i=0; i<nl; i++) {
    for (int j=0; j<nc; j++) {
      if(roi_data[j]>thres_preproc)
      {
        num_all_points++;
        cvSeqPush(all_points_seq, &cvPoint2D32f(j,i));
        //std::cout << "[j,i]: " << j << " " << i << " px+j-kernel_size/2: " << px+j-kernel_size/2 << " py+i-kernel_size/2: " << py+i-kernel_size/2 << "\n";
      }
      //else std::cout << "missing\n";
    }
    roi_data+= step;
  }

  //array storing all ellipse parameters found so far, and later the cluster they are put into
  CvMat* candidate_parameters = cvCreateMat( samples_num, 5, CV_32FC1 );
  CvMat* clusters = cvCreateMat( samples_num, 1, CV_32SC1 );

  //amount of all ellipse parameters found so far
  int ellipses_count = 0;

  CvRNG rng = cvRNG(cvGetTickCount());

  int samples_count=0;
  while(samples_count<samples_num)
  {

    bool points_ok = false;
    CvPoint2D32f* p[3];
    MyLine l[3];

    while(!points_ok)
    {
      int s[3];

      s[0]=0;
      s[1]=0;
      s[2]=0;

      //choose randomly 3 points from the array
      while(s[0]==s[1] || s[0]==s[2] || s[1]==s[2])
      {
        s[0] = cvRandInt(&rng) % all_points_seq->total;
        s[1] = cvRandInt(&rng) % all_points_seq->total;
        s[2] = cvRandInt(&rng) % all_points_seq->total;
        //std::cout << "\ns1: " << s[0] << " s2: " << s[1] << "\n";
      }

      points_ok = true;

      //if some point lies too near to the border, choose a new triple
      for(int c=0;c<3;c++)
      {
        p[c] = CV_GET_SEQ_ELEM( CvPoint2D32f, all_points_seq, s[c] );
        if(p[c]->x<kernel_size/2 || p[c]->x>(roi_size - kernel_size/2) || 
          p[c]->y<kernel_size/2 || p[c]->y>(roi_size - kernel_size/2)) 
        {
          points_ok=false; 
          break;
        }
      }

      //choose new points
      if(!points_ok) continue;

      //if some points lie too near or too far from each other, choose a new triple
      for(int c=0;c<3;c++)
      {
        int d = (c+1)%3;

        double a = p[c]->x - p[d]->x;
        double b = p[c]->y - p[d]->y;
        //std::cout << "\na: " << a << " b: " << b << "\n";

        double dist = sqrt(a*a + b*b);

        if(dist < min_dist || dist > max_dist) 
        {
          points_ok=false; 
          break;
        }
      }
    }

    //fit lines to chosen points
    bool all_lines_ok = true;
    for(int c=0;c<3;c++)
    {
      //std::cout << "\nx: " << p[c]->x << " y: " << p[c]->y << "\n";

      IplImage* kernel = cvCreateImage( cvSize(kernel_size,kernel_size), 8, 1 );
      cvSetImageROI(roi,cvRect(cvRound(p[c]->x-kernel_size/2),cvRound(p[c]->y-kernel_size/2),kernel_size,kernel_size)); 
      cvCopy(roi,kernel); 
      cvResetImageROI(roi);

      CvMemStorage* kernel_storage = cvCreateMemStorage(0);
      CvSeq* kernel_point_seq = cvCreateSeq( CV_32FC2, sizeof(CvSeq), sizeof(CvPoint2D32f), kernel_storage );
      unsigned char *kernel_data= reinterpret_cast<unsigned char *>(kernel->imageData);

      int nl= kernel->height; // number of lines
      int nc= kernel->width;// total number of element per line
      int step= kernel->widthStep; // effective width

      int num_points_found = 0;
      for (int i=0; i<nl; i++) {
        for (int j=0; j<nc; j++) {
          if(kernel_data[j]>thres_preproc)
          {
            num_points_found++;
            cvSeqPush(kernel_point_seq, &cvPoint2D32f(p[c]->x+j-kernel_size/2,p[c]->y+i-kernel_size/2));
            //std::cout << "[j,i]: " << j << " " << i << " px+j-kernel_size/2: " << px+j-kernel_size/2 << " py+i-kernel_size/2: " << py+i-kernel_size/2 << "\n";
          }
          //else std::cout << "missing\n";
        }
        kernel_data+= step;  // next line
      }

      //fit line only if at least kernel_num_points+1 points are found in the kernel
      if(num_points_found>kernel_num_points)
      {
        float *fline = new float[4];
        cvFitLine(kernel_point_seq,CV_DIST_L2,0,0.001,0.001,fline);

        l[c].vx = fline[0];
        l[c].vy = fline[1];
        l[c].px = fline[2];
        l[c].py = fline[3];
        //std::cout << "line params: " << fline[0] << ":" << fline[1] << ":" << fline[2] << ":" << fline[3] << "\n";
      }
      else all_lines_ok = false;

      cvClearMemStorage( kernel_storage ); 
      cvReleaseImage(&kernel);

      //std::stringstream filename;
      //filename << argv[2] << p << ".jpg";
      //std::cout << filename.str().c_str() << "\n\n";
      //cvSaveImage( filename.str().c_str(), kernel );

    }

    //if some line could not be fitted, choose new points
    if(!all_lines_ok) continue;


    //Find points M and T, which define the line on which the ellipse center has to lie
    CvPoint2D32f m[2], t[2], tm[2];
    for(int c=0;c<2;c++)
    {
      int d=c+1;
      m[c].x = (float) 0.5*(l[c].px + l[d].px);
      m[c].y = (float) 0.5*(l[c].py + l[d].py);

      t[c] = intersect( &l[c], &l[d]);

      tm[c].x = m[c].x - t[c].x;
      tm[c].y = m[c].y - t[c].y;

      //draw M, T and the lines related
      if(verbosity == 2)
      {
        float ray_vote = max_dist/2;
        CvPoint e;

        double e_length = ray_vote/sqrt((float)(tm[c].x*tm[c].x + tm[c].y*tm[c].y));
        e.x = cvRound(m[c].x + tm[c].x*e_length);
        e.y = cvRound(m[c].y + tm[c].y*e_length);

        cvCircle( color_roi, cvPoint(cvRound(m[c].x),cvRound(m[c].y)), 1, CV_RGB(0,255,0), -1, 8, 0 );
        cvCircle( color_roi, cvPoint(cvRound(t[c].x),cvRound(t[c].y)), 1, CV_RGB(0,0,255), -1, 8, 0 );
        cvLine( color_roi, cvPoint(cvRound(l[c].px),cvRound(l[c].py)), cvPoint(cvRound(l[d].px),cvRound(l[d].py)), CV_RGB(128,0,0), 1, 8 );
        cvLine( color_roi, cvPoint(cvRound(l[c].px),cvRound(l[c].py)), cvPoint(cvRound(t[c].x),cvRound(t[c].y)), CV_RGB(255,0,0), 1, 8 );
        cvLine( color_roi, cvPoint(cvRound(l[d].px),cvRound(l[d].py)), cvPoint(cvRound(t[c].x),cvRound(t[c].y)), CV_RGB(255,0,0), 1, 8 );

        if(cvClipLine( cvSize(roi->width,roi->height), &cvPoint(cvRound(m[c].x),cvRound(m[c].y)), &e ))
        {
          cvLine( color_roi, cvPoint(cvRound(m[c].x),cvRound(m[c].y)), e, CV_RGB(0,255,0), 1, 8 );
          //std::cout << " ";// << m[c].x << " " << m[c].y << " " << e.x << " " << e.y << "i ";
        }
      }
    }

    //calculate the candidate center
    CvPoint2D32f o_f = intersect( &myLine(tm[0].x,tm[0].y,m[0].x,m[0].y),&myLine(tm[1].x,tm[1].y,m[1].x,m[1].y));
    CvPoint o;
    o.x = cvRound(o_f.x);
    o.y = cvRound(o_f.y);

    //if center is in image, compute ellipse parameters
    if(!(o.x < 0 || o.x > roi_size || o.y < 0 || o.y > roi_size))
    {
      //cvCircle( color_roi, o, 1, CV_RGB(0,255,0), -1, 8, 0 );

      float matrix[9];
      float sol[] = { 1,  1,  1};

      CvPoint2D32f op[3];
      for(int c=0;c<3;c++)
      {
        op[c].x = p[c]->x - o_f.x;
        op[c].y = p[c]->y - o_f.y;

        matrix[3*c + 0] = op[c].x*op[c].x;
        matrix[3*c + 1] = (float) 2.0*op[c].x*op[c].y;
        matrix[3*c + 2] = op[c].y*op[c].y;

        //std::cout << matrix[c][0] << " : " << matrix[c][1] << " : " << matrix[c][2] << "\n";
      }

      //Compute ellipse equation parameters from center and three sample points
      CvMat* X = cvCreateMat( 3, 1, CV_32FC1 );
      cvSolve( &cvMat(3, 3, CV_32FC1, matrix), &cvMat(3, 1, CV_32FC1, sol), X, CV_SVD );
      if((X->data.fl[0]*X->data.fl[2] - X->data.fl[1]*X->data.fl[1])>0)
      {
        //std::cout << "a: " << X->data.fl[0] << "b: " << X->data.fl[1] << "c: " << X->data.fl[2] << "\n";

        if(verbosity == 2 || verbosity == 3)
          for(int c=0;c<3;c++)
          {
            ////parameters are normalized, so ellipse equation has to be = 1
            //std::cout << "ellipse equation1 (=1): " <<  X->data.fl[0]*op[c].x*op[c].x + 2.0*X->data.fl[1]*op[c].x*op[c].y + X->data.fl[2]*op[c].y*op[c].y << "\n";

            //show sample points
            cvCircle( color_roi, cvPoint(cvRound(p[c]->x),cvRound(p[c]->y)), 1, CV_RGB(0,255,0), -1, 8, 0 );
          }

          //compute geometric ellipse parameters from equation parameters
          float B = X->data.fl[2]/X->data.fl[0];
          float C = (float) -1.0/X->data.fl[0];
          float D = X->data.fl[1]/X->data.fl[0];

          float r1 = (float) sqrt(-2.0*C/((B+1) - sqrt((B-1)*(B-1) + 4.0*D*D) ));
          float r2 = (float) sqrt(-2.0*C/((B+1) + sqrt((B-1)*(B-1) + 4.0*D*D) ));
          float theta_arc = (float) 0.5*atan(2.0*D/(1.0-B));
          float theta = (float) 90.0 - theta_arc*180.0/PI;

          //show candidate center
          if(verbosity == 3) cvCircle( color_roi, o, 1, CV_RGB(255,0,0), -1, 8, 0 );

          //determine order of the axes (this should be possible in the computation above..)
          float test_A = r1*r1*cos(theta_arc)*cos(theta_arc) + r2*r2*sin(theta_arc)*sin(theta_arc);
          float test_F = r1*r1*r2*r2;
          float test_AF = test_A/test_F;

          if(abs(X->data.fl[0] - test_AF)>0.0001)
            theta -= 90;

          //std::cout << "r1: " << r1 << "r2: " << r2 << "theta: " << theta << "\n";
          //std::cout << "A: " << test_A << "\n";
          //std::cout << "F: " << test_F << "\n";
          //std::cout << "A/F: " << test_AF << "\n"; 

          //if ellipse is not too big nor too small, add it
          if(r1<max_dist && r2<max_dist)
          {
            if(verbosity == 3) cvEllipse( color_roi, o, cvSize(cvRound(r1),cvRound(r2)), theta, 0, 360, CV_RGB(255,0,0), 1, 8, 0 );
            candidate_parameters->data.fl[ellipses_count*5+0] = (float) o.x;
            candidate_parameters->data.fl[ellipses_count*5+1] = (float) o.y;
            candidate_parameters->data.fl[ellipses_count*5+2] = (float) r1;
            candidate_parameters->data.fl[ellipses_count*5+3] = (float) r2;
            candidate_parameters->data.fl[ellipses_count*5+4] = (float) theta;
            ellipses_count++;
          }
      }

    }
    samples_count++;
  }

  //if no ellipses are found, don't do clustering
  if(ellipses_count > 0)
  {
    //cluster the candidate parameters by k-means clustering
    CvMat sub_candidate_parameters, sub_clusters;
    cvGetRows( candidate_parameters, &sub_candidate_parameters, 0, ellipses_count);
    cvGetRows( clusters, &sub_clusters, 0, ellipses_count);
    cvKMeans2( &sub_candidate_parameters, cluster_count, &sub_clusters,
      cvTermCriteria( CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 100, 0.001 ));

    //show the ellipses
    if(verbosity == 1)
      for( int c = 0; c < ellipses_count; c++ )
      {
        CvPoint o;
        o.x = cvRound(candidate_parameters->data.fl[c*5+0]);
        o.y = cvRound(candidate_parameters->data.fl[c*5+1]);
        int r1 = cvRound(candidate_parameters->data.fl[c*5+2]);
        int r2 = cvRound(candidate_parameters->data.fl[c*5+3]);
        float theta = candidate_parameters->data.fl[c*5+4];
        int cluster_idx = clusters->data.i[c];
        cvEllipse( color_roi, o, cvSize(r1,r2), theta, 0, 360, color_tab[cluster_idx], 1, 8, 0 );
      }
  }
  cvSaveImage( argv[2], color_roi );
  
}
