/******************************************************************************
*
*   This file is part of the software tool "Cervix Image Processing 2 (CIP2)".
*
*   Author: Christoph Bergmeir
*   Supervisor: José M. Benítez
*   Copyright (c) DiCITS Lab, Sci2s group, DECSAI, University of Granada.
*
*   This program is free software; you can redistribute it and/or
*   modify it under the terms of the GNU General Public License
*   as published by the Free Software Foundation; either version 2
*   of the License, or (at your option) any later version.
*   
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
*   MA  02110-1301, USA.
*
******************************************************************************/

#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"

#include "itkConnectedComponentImageFilter.h"
#include "itkRelabelComponentImageFilter.h"
#include "itkThresholdImageFilter.h"

int main( int argc, char * argv[] )
{
  if( argc < 3 )
    {
    std::cerr << "Usage: " << std::endl;
    std::cerr << argv[0] << "  inputImageFile   outputImageFile   maskSizeX   maskSizeY" << std::endl;
    return EXIT_FAILURE;
    }

  typedef   unsigned char  InputPixelType;
  typedef   unsigned char  OutputPixelType;

  typedef itk::Image< InputPixelType,  2 >   InputImageType;
  typedef itk::Image< OutputPixelType, 2 >   OutputImageType;

  typedef itk::ImageFileReader< InputImageType  >  ReaderType;
  typedef itk::ImageFileWriter< OutputImageType >  WriterType;

  ReaderType::Pointer reader = ReaderType::New();
  WriterType::Pointer writer = WriterType::New();

  reader->SetFileName( argv[1] );
  writer->SetFileName( argv[2] );

  typedef itk::ConnectedComponentImageFilter< InputImageType, OutputImageType > ConnectedComponentFilterType;
  typedef itk::RelabelComponentImageFilter< InputImageType, OutputImageType > RelabelComponentFilterType;
  typedef itk::ThresholdImageFilter< InputImageType >  ThresholdFilterType;

  ConnectedComponentFilterType::Pointer connectedComponentFilter = ConnectedComponentFilterType::New();
  RelabelComponentFilterType::Pointer relabelComponentFilter = RelabelComponentFilterType::New();
  ThresholdFilterType::Pointer thresholdFilter = ThresholdFilterType::New();

  int r_min = atoi(argv[3]); // radius along x
  int r_max = atoi(argv[4]); // radius along y

  unsigned long area_min = r_min * r_min * 3.1415;
  unsigned long area_max = r_max * r_max * 3.1415;

  connectedComponentFilter->SetInput( reader->GetOutput() );

  relabelComponentFilter->SetInput( connectedComponentFilter->GetOutput() );
  relabelComponentFilter->SetMinimumObjectSize(area_min);
  relabelComponentFilter->Update();

  unsigned long objectNum = relabelComponentFilter->GetNumberOfObjects();
  std::cout << "Number of Objects: " << objectNum << "\n";
  std::cout << "OrigNr of Objects: " << relabelComponentFilter->GetOriginalNumberOfObjects() << "\n";


  std::vector<unsigned long> objectSize = relabelComponentFilter->GetSizeOfObjectsInPixels();
  unsigned long label_start = 1;

  for(unsigned long x=0;x<objectNum;x++)
  {
    if(objectSize[x]>area_max) 
    {
      std::cout << "too big ";
      label_start++;
    }
    std::cout << "Size of Object " << x << ": " << objectSize[x] << "\n";
  }

  thresholdFilter->SetInput( relabelComponentFilter->GetOutput() );
  thresholdFilter->SetOutsideValue( 0 );
  thresholdFilter->ThresholdBelow( label_start );
  //thresholdFilter->Update();

  writer->SetInput( thresholdFilter->GetOutput() );
  writer->Update();

  return EXIT_SUCCESS;
}

