/******************************************************************************
*
*   This file is part of the software tool "Cervix Image Processing 2 (CIP2)".
*
*   Author: Christoph Bergmeir
*   Supervisor: José M. Benítez
*   Copyright (c) DiCITS Lab, Sci2s group, DECSAI, University of Granada.
*
*   This program is free software; you can redistribute it and/or
*   modify it under the terms of the GNU General Public License
*   as published by the Free Software Foundation; either version 2
*   of the License, or (at your option) any later version.
*   
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
*   MA  02110-1301, USA.
*
******************************************************************************/

#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"

//#include "../Utilities/InsightJournal/binaryAttributeMorphology/itkShapeLabelObject.h"
//#include "../Utilities/InsightJournal/binaryAttributeMorphology/itkLabelMap.h"
//#include "../Utilities/InsightJournal/binaryAttributeMorphology/itkBinaryImageToShapeLabelMapFilter.h"
//#include "../Utilities/InsightJournal/binaryAttributeMorphology/itkLabelMapToBinaryImageFilter.h"

//this should work from ITK 3.16 on...
#include "itkShapeLabelObject.h"
#include "itkLabelMap.h"
#include "itkBinaryImageToShapeLabelMapFilter.h"
#include "itkLabelMapToBinaryImageFilter.h"

int main(int argc, char * argv[])
{
  const int dim = 2;
  typedef unsigned char PixelType;
  typedef itk::Image< PixelType, dim >    ImageType;

  //if( argc != 3)
  //  {
  //  std::cerr << "usage: " << argv[0] << " input foreground" << std::endl;
  //  // std::cerr << "  : " << std::endl;
  //  exit(1);
  //  }

  // read the input image
  typedef itk::ImageFileReader< ImageType > ReaderType;
  typedef itk::ImageFileWriter< ImageType >  WriterType;

  ReaderType::Pointer reader = ReaderType::New();
  WriterType::Pointer writer = WriterType::New();

  reader->SetFileName( argv[1] );
  writer->SetFileName( argv[2] );

  // define the object type. Here the ShapeLabelObject type
  // is chosen in order to read some attribute related to the shape
  // of the objects (by opposition to the content of the object, with
  // the StatisticsLabelObejct).
  typedef unsigned long LabelType;
  typedef itk::ShapeLabelObject< LabelType, dim > LabelObjectType;
  typedef itk::LabelMap< LabelObjectType > LabelMapType;

  // convert the image in a collection of objects
  typedef itk::BinaryImageToShapeLabelMapFilter< ImageType, LabelMapType > ConverterType;
  ConverterType::Pointer converter = ConverterType::New();

  converter->ComputePerimeterOn();

  typedef itk::LabelMapToBinaryImageFilter< LabelMapType, ImageType > ReConverterType;
  ReConverterType::Pointer reConverter = ReConverterType::New();

  converter->SetInput( reader->GetOutput() );
  converter->SetInputForegroundValue( atoi(argv[3]) );

  int r_min = atoi(argv[4]);
  int r_max = atoi(argv[5]);
  //float roundness = atof(argv[6]);

  unsigned long area_min = r_min * r_min * 3.1415;
  unsigned long area_max = r_max * r_max * 3.1415;

  // update the shape filter, so its output will be up to date
  converter->UpdateLargestPossibleRegion();

  // then we can read the attribute values we're interested in. The BinaryImageToShapeLabelMapFilter
  // produce consecutive labels, so we can use a for loop and GetLabelObject() method to retrieve
  // the label objects. If the labels are not consecutive, the GetNthLabelObject() method must be
  // use instead of GetLabelObject(), or an iterator on the label object container of the label map.
  LabelMapType::Pointer labelMap = converter->GetOutput();
    //labelMap->UpdateLargestPossibleRegion();

  unsigned int num_labels = labelMap->GetNumberOfLabelObjects();

  for( unsigned int label=1; label<=num_labels; label++ )
    {
      LabelObjectType * labelObject = labelMap->GetLabelObject( label );

      //std::cout << label << "\t" << labelObject->GetPhysicalSize()
      //          << "\t" << labelObject->GetCentroid() << "\t"
      //          << labelObject->GetRoundness() << "\t" << labelMap->GetNumberOfLabelObjects();

      if(labelObject->GetPhysicalSize()<area_min || labelObject->GetPhysicalSize()>area_max || labelObject->GetRoundness()<1.35)
      {
        labelMap->RemoveLabel(label);
        //std::cout << "\t" << "removed ";
      }

      //std::cout << std::endl;

    }

  reConverter->SetInput(labelMap);
  writer->SetInput( reConverter->GetOutput() );
  writer->Update();

  return 0;
}

