/******************************************************************************
*
*   This file is part of the software tool "Cervix Image Processing 2 (CIP2)".
*
*   Author: Christoph Bergmeir
*   Supervisor: José M. Benítez
*   Copyright (c) DiCITS Lab, Sci2s group, DECSAI, University of Granada.
*
*   This program is free software; you can redistribute it and/or
*   modify it under the terms of the GNU General Public License
*   as published by the Free Software Foundation; either version 2
*   of the License, or (at your option) any later version.
*   
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*   
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
*   MA  02110-1301, USA.
*
******************************************************************************/

#include "itkOtsuMultipleThresholdsCalculator.h"

#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkScalarImageToHistogramGenerator.h"
#include "itkBinaryThresholdImageFilter.h"
#include "itkNumericTraits.h"

#include <stdio.h>
int main(int argc, char * argv[])
{
  typedef unsigned short InputPixelType;
  typedef unsigned char OutputPixelType;

  typedef itk::Image<InputPixelType, 2> InputImageType;
  typedef itk::Image<OutputPixelType, 2> OutputImageType;

  typedef itk::Statistics::ScalarImageToHistogramGenerator<InputImageType>
      ScalarImageToHistogramGeneratorType;

  typedef ScalarImageToHistogramGeneratorType::HistogramType HistogramType;

  typedef itk::OtsuMultipleThresholdsCalculator<HistogramType> CalculatorType;

  typedef itk::ImageFileReader<InputImageType> ReaderType;
  typedef itk::ImageFileWriter<OutputImageType> WriterType;

  typedef itk::BinaryThresholdImageFilter<InputImageType, OutputImageType>
      FilterType;

  ScalarImageToHistogramGeneratorType::Pointer scalarImageToHistogramGenerator =
      ScalarImageToHistogramGeneratorType::New();

  CalculatorType::Pointer calculator = CalculatorType::New();
  FilterType::Pointer filter = FilterType::New();

  ReaderType::Pointer reader = ReaderType::New();
  WriterType::Pointer writer = WriterType::New();

  scalarImageToHistogramGenerator->SetNumberOfBins(128);
  calculator->SetNumberOfThresholds(atoi(argv[3]));

  const OutputPixelType outsideValue = 0;
  const OutputPixelType insideValue = 255;

  filter->SetOutsideValue(outsideValue);
  filter->SetInsideValue(insideValue);

  reader->SetFileName(argv[1]);

  scalarImageToHistogramGenerator->SetInput(reader->GetOutput());
  calculator->SetInputHistogram(scalarImageToHistogramGenerator->GetOutput());
  filter->SetInput(reader->GetOutput());
  writer->SetInput(filter->GetOutput());

  //Invoke pipeline
  try
  {
    reader->Update();
  } catch (itk::ExceptionObject & excp)
  {
    std::cerr << "Exception thrown while reading image" << excp << std::endl;
  }
  scalarImageToHistogramGenerator->Compute();

  try
  {
    calculator->Update();
  } catch (itk::ExceptionObject & excp)
  {
    std::cerr << "Exception thrown " << excp << std::endl;
  }

  const CalculatorType::OutputType &thresholdVector = calculator->GetOutput();
  CalculatorType::OutputType::const_iterator itNum = thresholdVector.begin();

  InputPixelType lowerThreshold = 0;
  InputPixelType upperThreshold;

  std::string outputFileBase = argv[2];
  char outputFilename[1000];
  std::string outputFile = outputFileBase + ".%03d.jpg";

  for (; itNum < thresholdVector.end(); itNum++)
  {

    upperThreshold = static_cast<InputPixelType> (*itNum);

    filter->SetLowerThreshold(lowerThreshold);
    filter->SetUpperThreshold(upperThreshold);

    lowerThreshold = upperThreshold;

    std::cout << "Threshold: " << upperThreshold << "\n";

    sprintf(outputFilename, outputFile.c_str(), (itNum
        - thresholdVector.begin()));
    //std::cout << outputFilename << "\n";
    writer->SetFileName(outputFilename);

    try
    {
      writer->Update();
    } catch (itk::ExceptionObject & excp)
    {
      std::cerr << "Exception thrown " << excp << std::endl;
    }
  }

  return EXIT_SUCCESS;
}

